"""see upath/tests/conftest.py for fixtures"""

import sys
import warnings

import fsspec
import pytest

from upath import UPath
from upath.implementations.cloud import S3Path

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import extends_base
from ..utils import overrides_base
from ..utils import posixify


def silence_botocore_datetime_deprecation(cls):
    # botocore uses datetime.datetime.utcnow in 3.12 which is deprecated
    # see: https://github.com/boto/boto3/issues/3889#issuecomment-1751296363
    if sys.version_info >= (3, 12):
        return pytest.mark.filterwarnings(
            "ignore"
            r":datetime.datetime.utcnow\(\) is deprecated"
            ":DeprecationWarning"
        )(cls)
    else:
        return cls


@silence_botocore_datetime_deprecation
class TestUPathS3(BaseTests, metaclass=OverrideMeta):
    SUPPORTS_EMPTY_DIRS = False

    @pytest.fixture(autouse=True)
    def path(self, s3_fixture):
        path, anon, s3so = s3_fixture
        self.path = UPath(path, anon=anon, **s3so)
        self.anon = anon
        self.s3so = s3so

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, S3Path)

    @extends_base
    def test_rmdir(self):
        dirname = "rmdir_test"
        mock_dir = self.path.joinpath(dirname)
        mock_dir.joinpath("test.txt").touch()
        mock_dir.rmdir()
        assert not mock_dir.exists()
        with pytest.raises(NotADirectoryError):
            self.path.joinpath("file1.txt").rmdir()

    @extends_base
    def test_relative_to_extra(self):
        assert "file.txt" == str(
            UPath("s3://test_bucket/file.txt").relative_to(UPath("s3://test_bucket"))
        )

    @extends_base
    def test_iterdir_root(self):
        client_kwargs = self.path.storage_options["client_kwargs"]
        bucket_path = UPath("s3://other_test_bucket", client_kwargs=client_kwargs)
        bucket_path.mkdir()

        (bucket_path / "test1.txt").touch()
        (bucket_path / "test2.txt").touch()

        for x in bucket_path.iterdir():
            assert x.name != ""
            assert x.exists()

    @extends_base
    @pytest.mark.parametrize(
        "joiner", [["bucket", "path", "file"], ["bucket/path/file"]]
    )
    def test_no_bucket_joinpath(self, joiner):
        path = UPath("s3://", anon=self.anon, **self.s3so)
        path = path.joinpath(*joiner)
        assert str(path) == "s3://bucket/path/file"

    @extends_base
    def test_creating_s3path_with_bucket(self):
        path = UPath("s3://", bucket="bucket", anon=self.anon, **self.s3so)
        assert str(path) == "s3://bucket/"

    @extends_base
    def test_iterdir_with_plus_in_name(self, s3_with_plus_chr_name):
        bucket, anon, s3so = s3_with_plus_chr_name
        p = UPath(
            f"s3://{bucket}/manual__2022-02-19T14:31:25.891270+00:00",
            anon=True,
            **s3so,
        )

        files = list(p.iterdir())
        assert len(files) == 1
        (file,) = files
        assert file == p.joinpath("file.txt")

    @extends_base
    @pytest.mark.xfail(reason="fsspec/universal_pathlib#144")
    def test_rglob_with_double_fwd_slash(self, s3_with_double_fwd_slash_files):
        import boto3
        import botocore.exceptions

        bucket, anon, s3so = s3_with_double_fwd_slash_files

        conn = boto3.resource("s3", **s3so["client_kwargs"])
        # ensure there's no s3://bucket/key.txt object
        with pytest.raises(botocore.exceptions.ClientError, match=".*Not Found.*"):
            conn.Object(bucket, "key.txt").load()
        # ensure there's a s3://bucket//key.txt object
        assert conn.Object(bucket, "/key.txt").get()["Body"].read() == b"hello world"

        p0 = UPath(f"s3://{bucket}//key.txt", **s3so)
        assert p0.read_bytes() == b"hello world"
        p1 = UPath(f"s3://{bucket}", **s3so)
        assert list(p1.rglob("*.txt")) == [p0]


@pytest.fixture
def s3_with_plus_chr_name(s3_server):
    anon, s3so = s3_server
    s3 = fsspec.filesystem("s3", anon=False, **s3so)
    bucket = "plus_chr_bucket"
    path = f"{bucket}/manual__2022-02-19T14:31:25.891270+00:00"
    s3.mkdir(path)
    s3.touch(f"{path}/file.txt")
    s3.invalidate_cache()
    try:
        yield bucket, anon, s3so
    finally:
        if s3.exists(bucket):
            for dir, _, keys in s3.walk(bucket):
                for key in keys:
                    if key.rstrip("/"):
                        s3.rm(f"{dir}/{key}")


@pytest.fixture
def s3_with_double_fwd_slash_files(s3_server):
    anon, s3so = s3_server
    s3 = fsspec.filesystem("s3", anon=False, **s3so)
    bucket = "double_fwd_slash_bucket"
    s3.mkdir(bucket + "/")
    s3.pipe_file(f"{bucket}//key.txt", b"hello world")
    try:
        yield bucket, anon, s3so
    finally:
        if s3.exists(bucket):
            for dir, _, keys in s3.walk(bucket):
                for key in keys:
                    if key.rstrip("/"):
                        s3.rm(f"{dir}/{key}")


def test_path_with_hash_and_space():
    assert "with#hash and space" in UPath("s3://bucket/with#hash and space/abc").parts


def test_pathlib_consistent_join():
    b0 = UPath("s3://mybucket/withkey/").joinpath("subfolder/myfile.txt")
    b1 = UPath("s3://mybucket/withkey").joinpath("subfolder/myfile.txt")
    assert b0 == b1
    assert "s3://mybucket/withkey/subfolder/myfile.txt" == str(b0) == str(b1)


def test_copy__object_key_collides_with_dir_prefix(s3_server, tmp_path):
    anon, s3so = s3_server

    s3 = fsspec.filesystem("s3", anon=anon, **{**s3so, "use_listings_cache": False})
    bucket = "copy_into_collision_bucket"
    s3.mkdir(bucket + "/src" + "/common_prefix/")
    # object under common prefix as key
    s3.pipe_file(f"{bucket}/src/common_prefix", b"hello world")
    # store more objects with same prefix
    s3.pipe_file(f"{bucket}/src/common_prefix/file1.txt", b"1")
    s3.pipe_file(f"{bucket}/src/common_prefix/file2.txt", b"2")

    # make sure the sources have a collision
    assert s3.isdir(f"{bucket}/src/common_prefix")
    assert s3.isfile(f"{bucket}/src/common_prefix")
    assert s3.isfile(f"{bucket}/src/common_prefix/file1.txt")
    assert s3.isfile(f"{bucket}/src/common_prefix/file2.txt")
    # prepare source and destination
    src = UPath(f"s3://{bucket}/src", anon=anon, **s3so)
    dst = UPath(tmp_path)

    def on_collision_rename_file(src, dst):
        warnings.warn(
            f"{src!s} collides with prefix. Renaming target file object to {dst!s}",
            UserWarning,
            stacklevel=3,
        )
        return (
            dst.with_suffix(dst.suffix + ".COLLISION"),
            dst,
        )

    # perform copy
    src.copy_into(dst, on_name_collision=on_collision_rename_file)

    # check results
    dst_files = sorted(posixify(x.relative_to(tmp_path)) for x in dst.glob("**/*"))
    assert dst_files == [
        "src",
        "src/common_prefix",
        "src/common_prefix.COLLISION",
        "src/common_prefix/file1.txt",
        "src/common_prefix/file2.txt",
    ]



from pathlib import Path

import pytest

from upath import UPath

DRIVE_ROOT_ANCHOR_TESTS = [
    # cloud
    ("s3://bucket", "bucket", "/", "bucket/", ("bucket/",)),
    ("s3://bucket/a", "bucket", "/", "bucket/", ("bucket/", "a")),
    ("gs://bucket", "bucket", "/", "bucket/", ("bucket/",)),
    ("gs://bucket/a", "bucket", "/", "bucket/", ("bucket/", "a")),
    ("az://bucket", "bucket", "/", "bucket/", ("bucket/",)),
    ("az://bucket/a", "bucket", "/", "bucket/", ("bucket/", "a")),
    # data
    (
        "data:text/plain,A%20brief%20note",
        "",
        "",
        "",
        ("data:text/plain,A%20brief%20note",),
    ),
    # github
    ("github://user:token@repo/abc", "", "", "", ("abc",)),
    # hdfs
    ("hdfs://a/b/c", "", "/", "/", ("/", "b", "c")),
    ("hdfs:///a/b/c", "", "/", "/", ("/", "a", "b", "c")),
    # http
    ("http://a/", "http://a", "/", "http://a/", ("http://a/", "")),
    ("http://a/b/c", "http://a", "/", "http://a/", ("http://a/", "b", "c")),
    ("https://a/b/c", "https://a", "/", "https://a/", ("https://a/", "b", "c")),
    # memory
    ("memory://a/b/c", "", "/", "/", ("/", "a", "b", "c")),
    ("memory:///a/b/c", "", "/", "/", ("/", "a", "b", "c")),
    # sftp
    ("sftp://a/b/c", "", "/", "/", ("/", "b", "c")),
    ("sftp:///a/b/c", "", "/", "/", ("/", "a", "b", "c")),
    # smb
    ("smb://a/b/c", "", "/", "/", ("/", "b", "c")),
    ("smb:///a/b/c", "", "/", "/", ("/", "a", "b", "c")),
    # webdav
    ("webdav+http://host.com/a/b/c", "", "", "", ("a", "b", "c")),
    ("webdav+http://host.com/a/b/c", "", "", "", ("a", "b", "c")),
    # local
    (
        "file:///a/b/c",
        Path("/a/b/c").absolute().drive,
        Path("/").absolute().root.replace("\\", "/"),
        Path("/").absolute().anchor.replace("\\", "/"),
        tuple(x.replace("\\", "/") for x in Path("/a/b/c").absolute().parts),
    ),
]


@pytest.mark.parametrize(
    "path,drive,root,anchor",
    [x[0:4] for x in DRIVE_ROOT_ANCHOR_TESTS],
)
def test_drive_root_anchor(path, drive, root, anchor):
    p = UPath(path)
    assert (p.drive, p.root, p.anchor) == (drive, root, anchor)


@pytest.mark.parametrize(
    "path,parts",
    [(x[0], x[4]) for x in DRIVE_ROOT_ANCHOR_TESTS],
)
def test_parts(path, parts):
    p = UPath(path)
    assert p.parts == parts



import os
import zipfile

import pytest

from upath import UnsupportedOperation
from upath import UPath
from upath.implementations.zip import ZipPath

from ..cases import JoinablePathTests
from ..cases import NonWritablePathTests
from ..cases import ReadablePathTests
from ..utils import OverrideMeta
from ..utils import extends_base
from ..utils import overrides_base


@pytest.fixture(scope="function")
def zipped_testdir_file(local_testdir, tmp_path_factory):
    base = tmp_path_factory.mktemp("zippath")
    zip_path = base / "test.zip"
    with zipfile.ZipFile(zip_path, "w") as zf:
        for root, _, files in os.walk(local_testdir):
            for file in files:
                full_path = os.path.join(root, file)
                arcname = os.path.relpath(full_path, start=local_testdir)
                zf.write(full_path, arcname=arcname)
    return str(zip_path)


@pytest.fixture(scope="function")
def empty_zipped_testdir_file(tmp_path):
    tmp_path = tmp_path.joinpath("zippath")
    tmp_path.mkdir()
    zip_path = tmp_path / "test.zip"

    with zipfile.ZipFile(zip_path, "w"):
        pass
    return str(zip_path)


class TestZipPath(
    JoinablePathTests,
    ReadablePathTests,
    NonWritablePathTests,
    metaclass=OverrideMeta,
):
    @pytest.fixture(autouse=True)
    def path(self, zipped_testdir_file, request):
        try:
            (mode,) = request.param
        except (ValueError, TypeError, AttributeError):
            mode = "r"
        self.path = UPath("zip://", fo=zipped_testdir_file, mode=mode)
        try:
            yield
        finally:
            self.path.fs.clear_instance_cache()

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, ZipPath)

    @extends_base
    def test_write_mode_is_disabled(self, tmp_path):
        with pytest.raises(UnsupportedOperation):
            UPath("zip://", fo=tmp_path.joinpath("myzip.zip"), mode="a")
        with pytest.raises(UnsupportedOperation):
            UPath("zip://", fo=tmp_path.joinpath("myzip.zip"), mode="x")
        with pytest.raises(UnsupportedOperation):
            UPath("zip://", fo=tmp_path.joinpath("myzip.zip"), mode="w")


@pytest.fixture(scope="function")
def zipped_testdir_file_in_memory(zipped_testdir_file, clear_fsspec_memory_cache):
    p = UPath(zipped_testdir_file, protocol="file")
    t = p.move(UPath("memory:///myzipfile.zip"))
    assert t.protocol == "memory"
    assert t.exists()
    yield t.as_uri()


class TestChainedZipPath(TestZipPath):

    @pytest.fixture(autouse=True)
    def path(self, zipped_testdir_file_in_memory, request):
        try:
            (mode,) = request.param
        except (ValueError, TypeError, AttributeError):
            mode = "r"
        self.path = UPath(
            "zip://", fo="/myzipfile.zip", mode=mode, target_protocol="memory"
        )



import pytest

from upath import UPath
from upath.implementations.memory import MemoryPath

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import overrides_base


class TestMemoryPath(BaseTests, metaclass=OverrideMeta):
    @pytest.fixture(autouse=True)
    def path(self, local_testdir):
        if not local_testdir.startswith("/"):
            local_testdir = "/" + local_testdir
        path = f"memory:{local_testdir}"
        self.path = UPath(path)
        self.prepare_file_system()

    @overrides_base
    def test_is_correct_class(self):
        assert isinstance(self.path, MemoryPath)


@pytest.mark.parametrize(
    "path, expected",
    [
        ("memory:/", "memory://"),
        ("memory:/a", "memory://a"),
        ("memory:/a/b", "memory://a/b"),
        ("memory://", "memory://"),
        ("memory://a", "memory://a"),
        ("memory://a/b", "memory://a/b"),
        ("memory:///", "memory://"),
        ("memory:///a", "memory://a"),
        ("memory:///a/b", "memory://a/b"),
    ],
)
def test_string_representation(path, expected):
    path = UPath(path)
    assert str(path) == expected



"""Automation using nox."""

import glob
import os
import sys

import nox

nox.options.reuse_existing_virtualenvs = True
nox.options.error_on_external_run = True

nox.needs_version = ">=2024.04.15"
nox.options.default_venv_backend = "uv"

nox.options.sessions = "lint", "tests", "type-checking", "type-safety"
locations = ("upath",)
running_in_ci = os.environ.get("CI", "") != ""

SUPPORTED_PYTHONS = ["3.9", "3.10", "3.11", "3.12", "3.13", "3.14"]
BASE_PYTHON = SUPPORTED_PYTHONS[-3]
MIN_PYTHON = SUPPORTED_PYTHONS[0]


@(lambda f: f())
def FSSPEC_MIN_VERSION() -> str:
    """Get the minimum fsspec version boundary from pyproject.toml."""
    try:
        from packaging.requirements import Requirement

        if sys.version_info >= (3, 11):
            from tomllib import load as toml_load
        else:
            from tomli import load as toml_load
    except ImportError:
        raise RuntimeError(
            "We rely on nox>=2024.04.15 depending on `packaging` and `tomli/tomllib`."
            " Please report if you see this error."
        )

    with open("pyproject.toml", "rb") as f:
        pyproject_data = toml_load(f)

    for requirement in pyproject_data["project"]["dependencies"]:
        req = Requirement(requirement)
        if req.name == "fsspec":
            for specifier in req.specifier:
                if specifier.operator == ">=":
                    return str(specifier.version)
    raise RuntimeError("Could not find fsspec minimum version in pyproject.toml")


@nox.session(python=SUPPORTED_PYTHONS)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    # workaround in case no aiohttp binary wheels are available
    if session.python == "3.14":
        session.env["AIOHTTP_NO_EXTENSIONS"] = "1"
        session.install(".[tests,dev]", "pydantic>=2.12.0a1")
    else:
        session.install(".[tests,dev,dev-third-party]")
    session.run("uv", "pip", "freeze", silent=not running_in_ci)
    session.run(
        "pytest",
        "-m",
        "not hdfs",
        "--cov",
        "--cov-config=pyproject.toml",
        *session.posargs,
        env={"COVERAGE_FILE": f".coverage.{session.python}"},
    )


@nox.session(python=MIN_PYTHON, name="tests-minversion")
def tests_minversion(session: nox.Session) -> None:
    session.install(f"fsspec=={FSSPEC_MIN_VERSION}", ".[tests,dev]")
    session.run("uv", "pip", "freeze", silent=not running_in_ci)
    session.run(
        "pytest",
        "-m",
        "not hdfs",
        "--cov",
        "--cov-config=pyproject.toml",
        *session.posargs,
        env={"COVERAGE_FILE": f".coverage.{session.python}"},
    )


tests_minversion.__doc__ = f"Run the test suite with fsspec=={FSSPEC_MIN_VERSION}."


@nox.session
def lint(session: nox.Session) -> None:
    """Run pre-commit hooks."""
    session.install("pre-commit")
    session.install("-e", ".[tests]")

    args = *(session.posargs or ("--show-diff-on-failure",)), "--all-files"
    session.run("pre-commit", "run", *args)


@nox.session
def safety(session: nox.Session) -> None:
    """Scan dependencies for insecure packages."""
    session.install(".")
    session.install("safety")
    session.run("safety", "check", "--full-report")


@nox.session
def build(session: nox.Session) -> None:
    """Build sdists and wheels."""
    session.install("build", "setuptools", "twine")
    session.run("python", "-m", "build")
    dists = glob.glob("dist/*")
    session.run("twine", "check", *dists, silent=True)


@nox.session
def develop(session: nox.Session) -> None:
    """Sets up a python development environment for the project."""
    session.run("uv", "venv", external=True)


@nox.session(name="type-checking", python=BASE_PYTHON)
def type_checking(session):
    """Run mypy checks."""
    session.install("-e", ".[typechecking]")
    session.run("python", "-m", "mypy")


@nox.session(name="type-safety", python=SUPPORTED_PYTHONS)
def type_safety(session):
    """Run typesafety tests."""
    session.install("-e", ".[typechecking]")
    session.run(
        "python",
        "-m",
        "pytest",
        "-v",
        "-p",
        "pytest-mypy-plugins",
        "--mypy-pyproject-toml-file",
        "pyproject.toml",
        "typesafety",
        *session.posargs,
    )


@nox.session(name="flavours-upgrade-deps", python=BASE_PYTHON)
def upgrade_flavours(session):
    session.run("uvx", "pur", "-r", "dev/requirements.txt")


@nox.session(name="flavours-codegen", python=BASE_PYTHON)
def generate_flavours(session):
    session.install("-r", "dev/requirements.txt")
    with open("upath/_flavour_sources.py", "w") as target:
        session.run(
            "python",
            "dev/fsspec_inspector/generate_flavours.py",
            stdout=target,
            stderr=None,
        )


@nox.session(name="docs-build", python=BASE_PYTHON)
def docs_build(session):
    """Build the documentation in strict mode."""
    session.install("--group=docs", "-e", ".")
    session.run("mkdocs", "build")


@nox.session(name="docs-serve", python=BASE_PYTHON)
def docs_serve(session):
    """Serve the documentation with live reloading."""
    session.install("--group=docs", "-e", ".")
    session.run("mkdocs", "serve", "--no-strict")



from __future__ import annotations

from typing import TYPE_CHECKING

from upath.registry import get_upath_class

if TYPE_CHECKING:
    from upath import UPath


def __getattr__(name: str) -> type[UPath]:
    if name.startswith("_") and name.endswith("Path"):
        from upath import UPath

        protocol = name[1:-4].lower()
        cls = get_upath_class(protocol)
        if cls is None:
            raise RuntimeError(
                f"Could not find fsspec implementation for protocol {protocol!r}"
            )
        elif not issubclass(cls, UPath):
            raise RuntimeError(
                "UPath implementation not a subclass of upath.UPath, {cls!r}"
            )
        return cls
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import os
from datetime import datetime
from datetime import timezone

import pytest

import upath


@pytest.fixture
def pth_file(tmp_path):
    f = tmp_path.joinpath("abc.txt")
    f.write_bytes(b"a")
    p = upath.UPath(f"file://{f.absolute().as_posix()}")
    yield p


def test_stat_repr(pth_file):
    assert repr(pth_file.stat()).startswith("UPathStatResult")


def test_stat_as_info(pth_file):
    dct = pth_file.stat().as_info()
    assert dct["size"] == pth_file.stat().st_size


def test_stat_atime(pth_file):
    atime = pth_file.stat().st_atime
    assert isinstance(atime, (float, int))


@pytest.mark.xfail(reason="fsspec does not return 'atime'")
def test_stat_atime_value(pth_file):
    atime = pth_file.stat().st_atime
    assert atime > 0


def test_stat_mtime(pth_file):
    mtime = pth_file.stat().st_mtime
    assert isinstance(mtime, (float, int))


def test_stat_mtime_value(pth_file):
    mtime = pth_file.stat().st_mtime
    assert mtime > 0


def test_stat_ctime(pth_file):
    ctime = pth_file.stat().st_ctime
    assert isinstance(ctime, (float, int))


@pytest.mark.xfail(reason="fsspec returns 'created' but not 'ctime'")
def test_stat_ctime_value(pth_file):
    ctime = pth_file.stat().st_ctime
    assert ctime > 0


def test_stat_birthtime(pth_file):
    birthtime = pth_file.stat().st_birthtime
    assert isinstance(birthtime, (float, int))


def test_stat_birthtime_value(pth_file):
    birthtime = pth_file.stat().st_birthtime
    assert birthtime > 0


def test_stat_seq_interface(pth_file):
    assert len(tuple(pth_file.stat())) == os.stat_result.n_sequence_fields
    assert isinstance(pth_file.stat().index(0), int)
    assert isinstance(pth_file.stat().count(0), int)
    assert isinstance(pth_file.stat()[0], int)


def test_stat_warn_if_dict_interface(pth_file):
    with pytest.warns(DeprecationWarning):
        pth_file.stat().keys()

    with pytest.warns(DeprecationWarning):
        pth_file.stat().items()

    with pytest.warns(DeprecationWarning):
        pth_file.stat().values()

    with pytest.warns(DeprecationWarning):
        pth_file.stat().get("size")

    with pytest.warns(DeprecationWarning):
        pth_file.stat().copy()

    with pytest.warns(DeprecationWarning):
        _ = pth_file.stat()["size"]


@pytest.mark.parametrize(
    "timestamp",
    [
        10,
        datetime(1970, 1, 1, 0, 0, 10, tzinfo=timezone.utc),
        "1970-01-01T00:00:10Z",
        "1970-01-01T00:00:10+00:00",
    ],
)
def test_timestamps(timestamp):
    from upath._stat import UPathStatResult

    s = UPathStatResult(
        [0] * 10,
        {
            "ctime": timestamp,
            "atime": timestamp,
            "mtime": timestamp,
            "created": timestamp,
        },
    )
    assert s.st_atime == 10.0
    assert s.st_ctime == 10.0
    assert s.st_mtime == 10.0


def test_bad_timestamp():
    from upath._stat import UPathStatResult

    with (
        pytest.raises(TypeError),
        pytest.warns(RuntimeWarning, "universal_pathlib/issues"),
    ):
        s = UPathStatResult([0] * 10, {"ctime": "bad"})
        _ = s.st_ctime



import pytest

from upath import UPath

from ..cases import BaseTests
from ..utils import OverrideMeta
from ..utils import overrides_base
from ..utils import skip_on_windows


@skip_on_windows
class TestUPathSMB(BaseTests, metaclass=OverrideMeta):

    @pytest.fixture(autouse=True)
    def path(self, smb_fixture):
        self.path = UPath(smb_fixture)

    @overrides_base
    def test_is_correct_class(self):
        from upath.implementations.smb import SMBPath

        assert isinstance(self.path, SMBPath)

    @overrides_base
    @pytest.mark.parametrize(
        "pattern",
        (
            "*.txt",
            pytest.param(
                "*",
                marks=pytest.mark.xfail(
                    reason="SMBFileSystem.info appends '/' to dirs"
                ),
            ),
            "**/*.txt",
        ),
    )
    def test_glob(self, pathlib_base, pattern):
        super().test_glob(pathlib_base, pattern)



from __future__ import annotations

import sys
from types import MappingProxyType
from typing import TYPE_CHECKING

from upath.core import UPath
from upath.types import JoinablePathLike

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any
    from typing import Literal

    if sys.version_info >= (3, 11):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from fsspec import AbstractFileSystem

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import SimpleCacheStorageOptions


__all__ = ["SimpleCachePath"]


class SimpleCachePath(UPath):
    __slots__ = ()

    if TYPE_CHECKING:

        def __init__(
            self,
            *args: JoinablePathLike,
            protocol: Literal["simplecache"] | None = ...,
            chain_parser: FSSpecChainParser = ...,
            **storage_options: Unpack[SimpleCacheStorageOptions],
        ) -> None: ...

    @classmethod
    def _fs_factory(
        cls,
        urlpath: str,
        protocol: str,
        storage_options: Mapping[str, Any],
    ) -> AbstractFileSystem:
        so = dict(storage_options)
        so.pop("fo", None)
        return super()._fs_factory(
            urlpath,
            protocol,
            so,
        )

    @property
    def storage_options(self) -> Mapping[str, Any]:
        so = self._storage_options.copy()
        so.pop("fo", None)
        return MappingProxyType(so)



""" upath._flavour_sources

<experimental!>

Warning
-------
  Do not modify this file manually!
  It is generated by `dev/generate_flavours.py`

To be able to parse the different filesystem uri schemes, we need
the string parsing functionality each of the filesystem implementations.
In an attempt to support parsing uris without having to import the
specific filesystems, we extract the necessary subset of the
AbstractFileSystem classes and generate a new "flavour" class for
each of the known filesystems. This will allow us to provide a
`PurePath` equivalent `PureUPath` for each protocol in the future
without a direct dependency on the underlying filesystem package.

"""
#
# skipping protocols:
#   - blockcache
#   - cached
#   - dir
#   - filecache
# protocol import errors:
#   - gdrive (Please install gdrive_fs for access to Google Drive)
#   - generic (GenericFileSystem: '_strip_protocol' not a classmethod)
#   - pyscript (Install requests (cpython) or run in pyscript)
#   - tos (Install tosfs to access ByteDance volcano engine Tinder Object Storage)
#   - tosfs (Install tosfs to access ByteDance volcano engine Tinder Object Storage)
#
from __future__ import annotations

import logging
import os
import re
from pathlib import PurePath
from pathlib import PureWindowsPath
from typing import Any
from typing import Literal
from typing import cast
from urllib.parse import parse_qs
from urllib.parse import urlsplit

from fsspec.implementations.local import make_path_posix
from fsspec.utils import infer_storage_options
from fsspec.utils import stringify_path

__all__ = [
    "AbstractFileSystemFlavour",
    "FileSystemFlavourBase",
    "flavour_registry",
]

logger = logging.getLogger(__name__)
flavour_registry: dict[str, type[FileSystemFlavourBase]] = {}


class FileSystemFlavourBase:
    """base class for the fsspec flavours"""

    protocol: str | tuple[str, ...]
    root_marker: Literal["/", ""]
    sep: Literal["/"]

    @classmethod
    def _strip_protocol(cls, path):
        raise NotImplementedError

    @staticmethod
    def _get_kwargs_from_urls(path):
        raise NotImplementedError

    @classmethod
    def _parent(cls, path):
        raise NotImplementedError

    def __init_subclass__(cls: Any, **kwargs):
        if isinstance(cls.protocol, str):
            protocols = (cls.protocol,)
        else:
            protocols = tuple(cls.protocol)
        for protocol in protocols:
            if protocol in flavour_registry:
                raise ValueError(f"protocol {protocol!r} already registered")
            flavour_registry[protocol] = cls


class AbstractFileSystemFlavour(FileSystemFlavourBase):
    __orig_class__ = 'fsspec.spec.AbstractFileSystem'
    __orig_version__ = '2025.10.0'
    protocol: str | tuple[str, ...] = 'abstract'
    root_marker: Literal['', '/'] = ''
    sep: Literal['/'] = '/'

    @classmethod
    def _strip_protocol(cls, path):
        """Turn path from fully-qualified to file-system-specific

        May require FS-specific handling, e.g., for relative paths or links.
        """
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        path = stringify_path(path)
        protos = (cls.protocol,) if isinstance(cls.protocol, str) else cls.protocol
        for protocol in protos:
            if path.startswith(protocol + "://"):
                path = path[len(protocol) + 3 :]
            elif path.startswith(protocol + "::"):
                path = path[len(protocol) + 2 :]
        path = path.rstrip("/")
        # use of root_marker to make minimum required path, e.g., "/"
        return path or cls.root_marker

    @staticmethod
    def _get_kwargs_from_urls(path):
        """If kwargs can be encoded in the paths, extract them here

        This should happen before instantiation of the class; incoming paths
        then should be amended to strip the options in methods.

        Examples may look like an sftp path "sftp://user@host:/my/path", where
        the user and host should become kwargs and later get stripped.
        """
        # by default, nothing happens
        return {}

    @classmethod
    def _parent(cls, path):
        path = cls._strip_protocol(path)
        if "/" in path:
            parent = path.rsplit("/", 1)[0].lstrip(cls.root_marker)
            return cls.root_marker + parent
        else:
            return cls.root_marker


class AsyncFileSystemWrapperFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.asyn_wrapper.AsyncFileSystemWrapper'
    __orig_version__ = '2025.10.0'
    protocol = ('asyncwrapper', 'async_wrapper')
    root_marker = ''
    sep = '/'


class AsyncLocalFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'morefs.asyn_local.AsyncLocalFileSystem'
    __orig_version__ = '0.2.2'
    protocol = ()
    root_marker = '/'
    sep = '/'
    local_file = True

    @classmethod
    def _strip_protocol(cls, path):
        path = stringify_path(path)
        if path.startswith("file://"):
            path = path[7:]
        elif path.startswith("file:"):
            path = path[5:]
        elif path.startswith("local://"):
            path = path[8:]
        elif path.startswith("local:"):
            path = path[6:]

        path = make_path_posix(path)
        if os.sep != "/":
            # This code-path is a stripped down version of
            # > drive, path = ntpath.splitdrive(path)
            if path[1:2] == ":":
                # Absolute drive-letter path, e.g. X:\Windows
                # Relative path with drive, e.g. X:Windows
                drive, path = path[:2], path[2:]
            elif path[:2] == "//":
                # UNC drives, e.g. \\server\share or \\?\UNC\server\share
                # Device drives, e.g. \\.\device or \\?\device
                if (index1 := path.find("/", 2)) == -1 or (
                    index2 := path.find("/", index1 + 1)
                ) == -1:
                    drive, path = path, ""
                else:
                    drive, path = path[:index2], path[index2:]
            else:
                # Relative path, e.g. Windows
                drive = ""

            path = path.rstrip("/") or cls.root_marker
            return drive + path

        else:
            return path.rstrip("/") or cls.root_marker

    @classmethod
    def _parent(cls, path):
        path = cls._strip_protocol(path)
        if os.sep == "/":
            # posix native
            return path.rsplit("/", 1)[0] or "/"
        else:
            # NT
            path_ = path.rsplit("/", 1)[0]
            if len(path_) <= 3:
                if path_[1:2] == ":":
                    # nt root (something like c:/)
                    return path_[0] + ":/"
            # More cases may be required here
            return path_


class AzureBlobFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'adlfs.spec.AzureBlobFileSystem'
    __orig_version__ = '2025.8.0'
    protocol = ('abfs', 'az', 'abfss')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path: str):
        """
        Remove the protocol from the input path

        Parameters
        ----------
        path: str
            Path to remove the protocol from

        Returns
        -------
        str
            Returns a path without the protocol
        """
        if isinstance(path, list):  # type: ignore[unreachable]
            return [cls._strip_protocol(p) for p in path]  # type: ignore[unreachable]

        STORE_SUFFIX = ".dfs.core.windows.net"
        logger.debug(f"_strip_protocol for {path}")
        if not path.startswith(("abfs://", "az://", "abfss://")):
            path = path.lstrip("/")
            path = "abfs://" + path
        ops = infer_storage_options(path)
        if "username" in ops:
            if ops.get("username", None):
                ops["path"] = ops["username"] + ops["path"]
        # we need to make sure that the path retains
        # the format {host}/{path}
        # here host is the container_name
        elif ops.get("host", None):
            if (
                ops["host"].count(STORE_SUFFIX) == 0
            ):  # no store-suffix, so this is container-name
                ops["path"] = ops["host"] + ops["path"]
        url_query = ops.get("url_query")
        if url_query is not None:
            ops["path"] = f"{ops['path']}?{url_query}"

        logger.debug(f"_strip_protocol({path}) = {ops}")
        stripped_path = ops["path"].lstrip("/")
        return stripped_path

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        """Get the account_name from the urlpath and pass to storage_options"""
        ops = infer_storage_options(urlpath)
        out = {}
        host = ops.get("host", None)
        if host:
            match = re.match(
                r"(?P<account_name>.+)\.(dfs|blob)\.core\.windows\.net", host
            )
            if match:
                account_name = match.groupdict()["account_name"]
                out["account_name"] = account_name
        url_query = ops.get("url_query")
        if url_query is not None:
            from urllib.parse import parse_qs

            parsed = parse_qs(url_query)
            if "versionid" in parsed:
                out["version_aware"] = True
        return out


class AzureDatalakeFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'adlfs.gen1.AzureDatalakeFileSystem'
    __orig_version__ = '2025.8.0'
    protocol = ('adl',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        ops = infer_storage_options(path)
        return ops["path"]

    @staticmethod
    def _get_kwargs_from_urls(paths):
        """Get the store_name from the urlpath and pass to storage_options"""
        ops = infer_storage_options(paths)
        out = {}
        if ops.get("host", None):
            out["store_name"] = ops["host"]
        return out


class BoxFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'boxfs.boxfs.BoxFileSystem'
    __orig_version__ = '0.3.0'
    protocol = ('box',)
    root_marker = '/'
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path) -> str:
        path = super()._strip_protocol(path)
        path = path.replace("\\", "/")
        # Make all paths start with root marker
        if not path.startswith(cls.root_marker):
            path = cls.root_marker + path
        return path


class DaskWorkerFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.dask.DaskWorkerFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('dask',)
    root_marker = ''
    sep = '/'

    @staticmethod
    def _get_kwargs_from_urls(path):
        so = infer_storage_options(path)
        if "host" in so and "port" in so:
            return {"client": f"{so['host']}:{so['port']}"}
        else:
            return {}


class DataFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.data.DataFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('data',)
    root_marker = ''
    sep = ''  # type: ignore[assignment]
    altsep = ' '  # type: ignore[assignment]


class DatabricksFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.dbfs.DatabricksFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('dbfs',)
    root_marker = ''
    sep = '/'


class DictFSFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'morefs.dict.DictFS'
    __orig_version__ = '0.2.2'
    protocol = ('dictfs',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path: str) -> str:
        if path.startswith("dictfs://"):
            path = path[len("dictfs://") :]
        if "::" in path or "://" in path:
            return path.rstrip("/")
        path = path.lstrip("/").rstrip("/")
        return "/" + path if path else cls.root_marker


class DropboxDriveFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'dropboxdrivefs.core.DropboxDriveFileSystem'
    __orig_version__ = '1.4.1'
    protocol = ('dropbox',)
    root_marker = ''
    sep = '/'


class FTPFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.ftp.FTPFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('ftp',)
    root_marker = '/'
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        return "/" + infer_storage_options(path)["path"].lstrip("/").rstrip("/")

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        out = infer_storage_options(urlpath)
        out.pop("path", None)
        out.pop("protocol", None)
        return out


class GCSFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'gcsfs.core.GCSFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('gs', 'gcs')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        path = stringify_path(path)
        protos = (cls.protocol,) if isinstance(cls.protocol, str) else cls.protocol
        for protocol in protos:
            if path.startswith(protocol + "://"):
                path = path[len(protocol) + 3 :]
            elif path.startswith(protocol + "::"):
                path = path[len(protocol) + 2 :]
        # use of root_marker to make minimum required path, e.g., "/"
        return path or cls.root_marker

    @classmethod
    def _get_kwargs_from_urls(cls, path):
        _, _, generation = cls._split_path(path, version_aware=True)
        if generation is not None:
            return {"version_aware": True}
        return {}

    @classmethod
    def _split_path(cls, path, version_aware=False):
        """
        Normalise GCS path string into bucket and key.

        Parameters
        ----------
        path : string
            Input path, like `gcs://mybucket/path/to/file`.
            Path is of the form: '[gs|gcs://]bucket[/key][?querystring][#fragment]'

        GCS allows object generation (object version) to be specified in either
        the URL fragment or the `generation` query parameter. When provided,
        the fragment will take priority over the `generation` query paramenter.

        Returns
        -------
            (bucket, key, generation) tuple
        """
        path = cls._strip_protocol(path).lstrip("/")
        if "/" not in path:
            return path, "", None
        bucket, keypart = path.split("/", 1)
        key = keypart
        generation = None
        if version_aware:
            parts = urlsplit(keypart)
            try:
                if parts.fragment:
                    generation = parts.fragment
                elif parts.query:
                    parsed = parse_qs(parts.query)
                    if "generation" in parsed:
                        generation = parsed["generation"][0]
                # Sanity check whether this could be a valid generation ID. If
                # it is not, assume that # or ? characters are supposed to be
                # part of the object name.
                if generation is not None:
                    int(generation)
                    key = parts.path
            except ValueError:
                generation = None
        return (
            bucket,
            key,
            generation,
        )


class GistFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.gist.GistFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('gist',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        """
        Remove 'gist://' from the path, if present.
        """
        # The default infer_storage_options can handle gist://username:token@id/file
        # or gist://id/file, but let's ensure we handle a normal usage too.
        # We'll just strip the protocol prefix if it exists.
        path = infer_storage_options(path).get("path", path)
        return path.lstrip("/")

    @staticmethod
    def _get_kwargs_from_urls(path):
        """
        Parse 'gist://' style URLs into GistFileSystem constructor kwargs.
        For example:
          gist://:TOKEN@<gist_id>/file.txt
          gist://username:TOKEN@<gist_id>/file.txt
        """
        so = infer_storage_options(path)
        out = {}
        if "username" in so and so["username"]:
            out["username"] = so["username"]
        if "password" in so and so["password"]:
            out["token"] = so["password"]
        if "host" in so and so["host"]:
            # We interpret 'host' as the gist ID
            out["gist_id"] = so["host"]

        # Extract SHA and filename from path
        if "path" in so and so["path"]:
            path_parts = so["path"].rsplit("/", 2)[-2:]
            if len(path_parts) == 2:
                if path_parts[0]:  # SHA present
                    out["sha"] = path_parts[0]
                if path_parts[1]:  # filename also present
                    out["filenames"] = [path_parts[1]]

        return out


class GitFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.git.GitFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('git',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        path = super()._strip_protocol(path).lstrip("/")
        if ":" in path:
            path = path.split(":", 1)[1]
        if "@" in path:
            path = path.split("@", 1)[1]
        return path.lstrip("/")

    @staticmethod
    def _get_kwargs_from_urls(path):
        path = path.removeprefix("git://")
        out = {}
        if ":" in path:
            out["path"], path = path.split(":", 1)
        if "@" in path:
            out["ref"], path = path.split("@", 1)
        return out


class GithubFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.github.GithubFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('github',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        opts = infer_storage_options(path)
        if "username" not in opts:
            return super()._strip_protocol(path)
        return opts["path"].lstrip("/")

    @staticmethod
    def _get_kwargs_from_urls(path):
        opts = infer_storage_options(path)
        if "username" not in opts:
            return {}
        out = {"org": opts["username"], "repo": opts["password"]}
        if opts["host"]:
            out["sha"] = opts["host"]
        return out


class HTTPFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.http.HTTPFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('http', 'https')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        """For HTTP, we always want to keep the full URL"""
        return path

    @classmethod
    def _parent(cls, path):
        # override, since _strip_protocol is different for URLs
        par = super()._parent(path)
        if len(par) > 7:  # "http://..."
            return par
        return ""


class HadoopFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.arrow.HadoopFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('hdfs', 'arrow_hdfs')
    root_marker = '/'
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        ops = infer_storage_options(path)
        path = ops["path"]
        if path.startswith("//"):
            # special case for "hdfs://path" (without the triple slash)
            path = path[1:]
        return path

    @staticmethod
    def _get_kwargs_from_urls(path):
        ops = infer_storage_options(path)
        out = {}
        if ops.get("host", None):
            out["host"] = ops["host"]
        if ops.get("username", None):
            out["user"] = ops["username"]
        if ops.get("port", None):
            out["port"] = ops["port"]
        if ops.get("url_query", None):
            queries = parse_qs(ops["url_query"])
            if queries.get("replication", None):
                out["replication"] = int(queries["replication"][0])
        return out


class HfFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'huggingface_hub.hf_file_system.HfFileSystem'
    __orig_version__ = '1.4.1'
    protocol = ('hf',)
    root_marker = ''
    sep = '/'


class JupyterFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.jupyter.JupyterFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('jupyter', 'jlab')
    root_marker = ''
    sep = '/'


class LakeFSFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'lakefs_spec.spec.LakeFSFileSystem'
    __orig_version__ = '0.12.0'
    protocol = ('lakefs',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        """Copied verbatim from the base class, save for the slash rstrip."""
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        spath = super()._strip_protocol(path)
        if stringify_path(path).endswith("/"):
            return spath + "/"
        return spath


class LibArchiveFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.libarchive.LibArchiveFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('libarchive',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        # file paths are always relative to the archive root
        return super()._strip_protocol(path).lstrip("/")


class LocalFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.local.LocalFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('file', 'local')
    root_marker = '/'
    sep = '/'
    local_file = True

    @classmethod
    def _strip_protocol(cls, path):
        path = stringify_path(path)
        if path.startswith("file://"):
            path = path[7:]
        elif path.startswith("file:"):
            path = path[5:]
        elif path.startswith("local://"):
            path = path[8:]
        elif path.startswith("local:"):
            path = path[6:]

        path = make_path_posix(path)
        if os.sep != "/":
            # This code-path is a stripped down version of
            # > drive, path = ntpath.splitdrive(path)
            if path[1:2] == ":":
                # Absolute drive-letter path, e.g. X:\Windows
                # Relative path with drive, e.g. X:Windows
                drive, path = path[:2], path[2:]
            elif path[:2] == "//":
                # UNC drives, e.g. \\server\share or \\?\UNC\server\share
                # Device drives, e.g. \\.\device or \\?\device
                if (index1 := path.find("/", 2)) == -1 or (
                    index2 := path.find("/", index1 + 1)
                ) == -1:
                    drive, path = path, ""
                else:
                    drive, path = path[:index2], path[index2:]
            else:
                # Relative path, e.g. Windows
                drive = ""

            path = path.rstrip("/") or cls.root_marker
            return drive + path

        else:
            return path.rstrip("/") or cls.root_marker

    @classmethod
    def _parent(cls, path):
        path = cls._strip_protocol(path)
        if os.sep == "/":
            # posix native
            return path.rsplit("/", 1)[0] or "/"
        else:
            # NT
            path_ = path.rsplit("/", 1)[0]
            if len(path_) <= 3:
                if path_[1:2] == ":":
                    # nt root (something like c:/)
                    return path_[0] + ":/"
            # More cases may be required here
            return path_


class MemFSFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'morefs.memory.MemFS'
    __orig_version__ = '0.2.2'
    protocol = ('memfs',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        if path.startswith("memfs://"):
            path = path[len("memfs://") :]
        return MemoryFileSystemFlavour._strip_protocol(path)  # pylint: disable=protected-access


class MemoryFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.memory.MemoryFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('memory',)
    root_marker = '/'
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        if isinstance(path, PurePath):
            if isinstance(path, PureWindowsPath):
                return LocalFileSystemFlavour._strip_protocol(path)
            else:
                path = stringify_path(path)

        path = path.removeprefix("memory://")
        if "::" in path or "://" in path:
            return path.rstrip("/")
        path = path.lstrip("/").rstrip("/")
        return "/" + path if path else ""


class OCIFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'ocifs.core.OCIFileSystem'
    __orig_version__ = '1.3.4'
    protocol = ('oci', 'ocilake')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        path = stringify_path(path)
        stripped_path = super()._strip_protocol(path)
        if stripped_path == cls.root_marker and "@" in path:
            return "@" + path.rstrip("/").split("@", 1)[1]
        return stripped_path

    @classmethod
    def _parent(cls, path):
        path = cls._strip_protocol(path.rstrip("/"))
        if "/" in path:
            return cls.root_marker + path.rsplit("/", 1)[0]
        elif "@" in path:
            return cls.root_marker + "@" + path.split("@", 1)[1]
        else:
            raise ValueError(f"the following path does not specify a namespace: {path}")


class OSSFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'ossfs.core.OSSFileSystem'
    __orig_version__ = '2025.5.0'
    protocol = ('oss',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        """Turn path from fully-qualified to file-system-specifi
        Parameters
        ----------
        path : Union[str, List[str]]
            Input path, like
            `http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject`
            `oss://mybucket/myobject`
        Examples
        --------
        >>> _strip_protocol(
            "http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject"
            )
        ('/mybucket/myobject')
        >>> _strip_protocol(
            "oss://mybucket/myobject"
            )
        ('/mybucket/myobject')
        """
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        path_string = stringify_path(path)
        if path_string.startswith("oss://"):
            path_string = path_string[5:]

        parser_re = r"https?://(?P<endpoint>oss.+aliyuncs\.com)(?P<path>/.+)"
        matcher = re.compile(parser_re).match(path_string)
        if matcher:
            path_string = matcher["path"]
        return path_string or cls.root_marker


class OverlayFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'morefs.overlay.OverlayFileSystem'
    __orig_version__ = '0.2.2'
    protocol = ('overlayfs',)
    root_marker = ''
    sep = '/'


class ReferenceFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.reference.ReferenceFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('reference',)
    root_marker = ''
    sep = '/'


class S3FileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 's3fs.core.S3FileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('s3', 's3a')
    root_marker = ''
    sep = '/'

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        """
        When we have a urlpath that contains a ?versionId=

        Assume that we want to use version_aware mode for
        the filesystem.
        """
        from urllib.parse import urlsplit

        url_query = urlsplit(urlpath).query
        out = {}
        if url_query is not None:
            from urllib.parse import parse_qs

            parsed = parse_qs(url_query)
            if "versionId" in parsed:
                out["version_aware"] = True
        return out


class SFTPFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.sftp.SFTPFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('sftp', 'ssh')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        return infer_storage_options(path)["path"]

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        out = infer_storage_options(urlpath)
        out.pop("path", None)
        out.pop("protocol", None)
        return out


class SMBFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.smb.SMBFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('smb',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        return infer_storage_options(path)["path"]

    @staticmethod
    def _get_kwargs_from_urls(path):
        # smb://workgroup;user:password@host:port/share/folder/file.csv
        out = infer_storage_options(path)
        out.pop("path", None)
        out.pop("protocol", None)
        return out


class SimpleCacheFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.cached.SimpleCacheFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('simplecache',)
    root_marker = ''
    sep = '/'
    local_file = True


class TarFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.tar.TarFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('tar',)
    root_marker = ''
    sep = '/'


class WandbFSFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'wandbfs._wandbfs.WandbFS'
    __orig_version__ = '0.0.2'
    protocol = ('wandb',)
    root_marker = ''
    sep = '/'


class WebHDFSFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.webhdfs.WebHDFS'
    __orig_version__ = '2025.10.0'
    protocol = ('webhdfs', 'webHDFS')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        return infer_storage_options(path)["path"]

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        out = infer_storage_options(urlpath)
        out.pop("path", None)
        out.pop("protocol", None)
        if "username" in out:
            out["user"] = out.pop("username")
        return out


class WebdavFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'webdav4.fsspec.WebdavFileSystem'
    __orig_version__ = '0.10.0'
    protocol = ('webdav', 'dav')
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path: str) -> str:
        """Strips protocol from the given path, overriding for type-casting."""
        stripped = super()._strip_protocol(path)
        return cast(str, stripped)


class XRootDFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec_xrootd.xrootd.XRootDFileSystem'
    __orig_version__ = '0.5.1'
    protocol = ('root',)
    root_marker = '/'
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path: str | list[str]) -> Any:
        if isinstance(path, str):
            if path.startswith(cls.protocol):
                x = urlsplit(path); return (x.path + f'?{x.query}' if x.query else '').rstrip("/") or cls.root_marker
            # assume already stripped
            return path.rstrip("/") or cls.root_marker
        elif isinstance(path, list):
            return [cls._strip_protocol(item) for item in path]
        else:
            raise ValueError("Strip protocol not given string or list")

    @staticmethod
    def _get_kwargs_from_urls(u: str) -> dict[Any, Any]:
        url = urlsplit(u)
        # The hostid encapsulates user,pass,host,port in one string
        return {"hostid": url.netloc}


class ZipFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'fsspec.implementations.zip.ZipFileSystem'
    __orig_version__ = '2025.10.0'
    protocol = ('zip',)
    root_marker = ''
    sep = '/'

    @classmethod
    def _strip_protocol(cls, path):
        # zip file paths are always relative to the archive root
        return super()._strip_protocol(path).lstrip("/")


class _DVCFileSystemFlavour(AbstractFileSystemFlavour):
    __orig_class__ = 'dvc.fs.dvc._DVCFileSystem'
    __orig_version__ = '3.66.1'
    protocol = ('dvc',)
    root_marker = '/'
    sep = '/'



from __future__ import annotations

import sys
from collections.abc import Iterator
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any
from typing import overload

from upath import UnsupportedOperation
from upath._chain import DEFAULT_CHAIN_PARSER
from upath._flavour import upath_strip_protocol
from upath.core import UPath
from upath.types import JoinablePathLike
from upath.types import SupportsPathLike
from upath.types import WritablePath

if TYPE_CHECKING:
    from typing import Literal
    from typing import TypeVar

    if sys.version_info >= (3, 11):
        from typing import Self
        from typing import Unpack
    else:
        from typing_extensions import Self
        from typing_extensions import Unpack

    from upath._chain import FSSpecChainParser
    from upath.types.storage_options import AzureStorageOptions
    from upath.types.storage_options import GCSStorageOptions
    from upath.types.storage_options import HfStorageOptions
    from upath.types.storage_options import S3StorageOptions

    _WT = TypeVar("_WT", bound="WritablePath")

__all__ = [
    "CloudPath",
    "GCSPath",
    "S3Path",
    "AzurePath",
    "HfPath",
]


class CloudPath(UPath):
    __slots__ = ()

    @classmethod
    def _transform_init_args(
        cls,
        args: tuple[JoinablePathLike, ...],
        protocol: str,
        storage_options: dict[str, Any],
    ) -> tuple[tuple[JoinablePathLike, ...], str, dict[str, Any]]:
        for key in ["bucket", "netloc"]:
            bucket = storage_options.pop(key, None)
            if bucket:
                if str(args[0]).startswith("/"):
                    args = (f"{protocol}://{bucket}{args[0]}", *args[1:])
                else:
                    args0 = upath_strip_protocol(args[0])
                    args = (f"{protocol}://{bucket}/", args0, *args[1:])
                break
        return super()._transform_init_args(args, protocol, storage_options)

    @property
    def root(self) -> str:
        if self._relative_base is not None:
            return ""
        return self.parser.sep

    def __str__(self) -> str:
        path = super().__str__()
        if self._relative_base is None:
            drive = self.parser.splitdrive(path)[0]
            if drive and path == f"{self.protocol}://{drive}":
                return f"{path}{self.root}"
        return path

    @property
    def path(self) -> str:
        self_path = super().path.rstrip(self.parser.sep)
        if (
            self._relative_base is None
            and self_path
            and self.parser.sep not in self_path
        ):
            return self_path + self.root
        return self_path

    @property
    def parts(self) -> Sequence[str]:
        parts = super().parts
        if self._relative_base is None and len(parts) == 2 and not parts[1]:
            return parts[:1]
        return parts

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False
    ) -> None:
        if not parents and not exist_ok and self.exists():
            raise FileExistsError(self.path)
        super().mkdir(mode=mode, parents=parents, exist_ok=exist_ok)


class GCSPath(CloudPath):
    __slots__ = ()

    def __init__(
        self,
        *args: JoinablePathLike,
        protocol: Literal["gcs", "gs"] | None = None,
        chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
        **storage_options: Unpack[GCSStorageOptions],
    ) -> None:
        super().__init__(
            *args, protocol=protocol, chain_parser=chain_parser, **storage_options
        )
        if not self.drive and len(self.parts) > 1:
            raise ValueError("non key-like path provided (bucket/container missing)")

    def mkdir(
        self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False
    ) -> None:
        try:
            super().mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        except TypeError as err:
            if "unexpected keyword argument 'create_parents'" in str(err):
                self.fs.mkdir(self.path)

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        # required for gcsfs<2025.5.0, see: https://github.com/fsspec/gcsfs/pull/676
        path = self.path
        if len(path) > 1:
            path = path.removesuffix(self.root)
        return self.fs.exists(path)


class S3Path(CloudPath):
    __slots__ = ()

    def __init__(
        self,
        *args: JoinablePathLike,
        protocol: Literal["s3", "s3a"] | None = None,
        chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
        **storage_options: Unpack[S3StorageOptions],
    ) -> None:
        super().__init__(
            *args, protocol=protocol, chain_parser=chain_parser, **storage_options
        )
        if not self.drive and len(self.parts) > 1:
            raise ValueError("non key-like path provided (bucket/container missing)")

    @overload
    def copy(self, target: _WT, **kwargs: Any) -> _WT: ...

    @overload
    def copy(self, target: SupportsPathLike | str, **kwargs: Any) -> Self: ...

    def copy(self, target: _WT | SupportsPathLike | str, **kwargs: Any) -> _WT | UPath:
        """
        Recursively copy this file or directory tree to the given destination.
        """
        # to allow _copy_from to check if a path isfile AND isdir
        # we need to disable s3fs's dircache mechanism because it
        # currently implements a XOR relation the two for objects
        # ref: fsspec/s3fs#999
        sopts = dict(self.storage_options)
        sopts["use_listings_cache"] = False
        new_self = type(self)(
            self.path,
            protocol=self.protocol,  # type: ignore
            **sopts,
        )
        assert type(self) is type(new_self)
        return super(type(new_self), new_self).copy(target, **kwargs)


class AzurePath(CloudPath):
    __slots__ = ()

    def __init__(
        self,
        *args: JoinablePathLike,
        protocol: Literal["abfs", "abfss", "adl", "az"] | None = None,
        chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
        **storage_options: Unpack[AzureStorageOptions],
    ) -> None:
        super().__init__(
            *args, protocol=protocol, chain_parser=chain_parser, **storage_options
        )
        if not self.drive and len(self.parts) > 1:
            raise ValueError("non key-like path provided (bucket/container missing)")


class HfPath(CloudPath):
    __slots__ = ()

    def __init__(
        self,
        *args: JoinablePathLike,
        protocol: Literal["hf"] | None = None,
        chain_parser: FSSpecChainParser = DEFAULT_CHAIN_PARSER,
        **storage_options: Unpack[HfStorageOptions],
    ) -> None:
        super().__init__(
            *args, protocol=protocol, chain_parser=chain_parser, **storage_options
        )

    @property
    def root(self) -> str:
        return ""

    def iterdir(self) -> Iterator[Self]:
        try:
            yield from super().iterdir()
        except NotImplementedError:
            raise UnsupportedOperation

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        raise UnsupportedOperation

    def mkdir(
        self,
        mode: int = 0o777,
        parents: bool = False,
        exist_ok: bool = False,
    ) -> None:
        raise UnsupportedOperation

    def unlink(self, missing_ok: bool = False) -> None:
        raise UnsupportedOperation

    def write_bytes(self, data: bytes) -> int:
        raise UnsupportedOperation("DataPath does not support writing")

    def write_text(
        self,
        data: str,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> int:
        raise UnsupportedOperation("DataPath does not support writing")
